-- Callback for simple motif program
-- Chris Morgan 3/9/95
with Text_Io;

with X;
use X;

use Interfaces.C;

with Mandel_Control_Package;
use  Mandel_Control_Package;

with Constants;
use  Constants;

with Area_Queue_Package;
with New_Area_Type_Package;

package body Callbacks_Package is

   Colour_Name_String : constant STRING := "none DarkSlateBlue MediumSlateBlue LightSlateBlue " &
                                           "DarkSeaGreen MediumSeaGreen LightSeaGreen " &
                                           "DarkOrchid1 DarkOrchid2 DarkOrchid3 " &
"IndianRed1 IndianRed2 IndianRed3 HotPink1 red white black ";
            

--in spec...   type Simple_Colour_Name_Type is         (None,Red,Blue,Green,Yellow,White,Black); -- Keep these two types

   type Colours_Array_Type is array(Simple_Colour_Name_Type) of aliased X.Xlib.Xcolor;

   Colours_Initialised : Boolean := False;
   Colours_Array : Colours_Array_Type;

   procedure Loc_Set_Colour
     (W	  : in Xt.Intrinsic.Widget;
      Col : in Simple_Colour_Name_Type;
      Fg  : in Boolean		       := true)
   is
      Status : X.Xlib.Status;
      Start,Stop : Integer := 6;
      Dummy : Character := 'a';
      Loc_Col : Simple_Colour_Name_Type := DarkSlateBlue;
   begin
      if not Colours_Initialised then
	 Colours_Initialised := True;
	 while Start < Colour_Name_String'Length loop
	    while Stop <= Colour_Name_String'Length and then
	      Colour_Name_String(Stop) /= ' ' loop
	       Stop := Stop + 1;
	       if Stop = Colour_Name_String'Length then
		  exit;
	       end if;
	    end loop;
	    --	 Text_Io.Put(Colour_Name_String(Start .. Stop - 1));
	    --	 Text_Io.Flush;
	    Status := X.Xlib.
	      Xallocnamedcolor(Display => Xt.Intrinsic.Xtdisplay(w),
			       Colormap => X.Xlib.Defaultcolormapofscreen(Xt.Intrinsic.Xtscreen(My_Window)),
			       Color_Name => Interfaces.C.To_C(Colour_Name_String(Start .. Stop - 1)), 
			       Screen_Def_Return => Colours_Array(Loc_col)'Access,
			       Exact_Def_Return => Colours_Array(none)'Access
			       );
	    if Status = 0 then
	       Text_Io.Put_Line("Allocating " & Colour_Name_String(Start .. Stop - 1) & " failed");
	       Text_Io.Flush;
--	    else
	--       Text_Io.Put_Line("Allocating " & Colour_Name_String(Start .. Stop - 1) & " succeeded");
	  --     Text_Io.Flush;
	    end if;
	    Start := Stop + 1;
	    Stop := Stop + 1;
	    if Loc_Col /= Simple_Colour_Name_Type'Last then
	       Loc_Col := Simple_Colour_Name_Type'Succ(Loc_Col);
	    end if;
	 end loop;
      end if;
      if Fg then
	 X.Xlib.
	   Xsetforeground(Display    => Xt.Intrinsic.Xtdisplay(w),
			  Gc         => The_Gc,
			  Foreground => Colours_Array(Col).Pixel);
      else
	 X.Xlib.
	   Xsetbackground(Display    => Xt.Intrinsic.Xtdisplay(w),
			  Gc         => The_Gc,
			  background => Colours_Array(Col).Pixel);
      end if;
   end;

   procedure Blackout ( W : Xt.Intrinsic.Widget ) is
   begin
      X.Xlib.XClearWindow(Xt.Intrinsic.Xtdisplay(Drawing_Area),
			  Xt.Intrinsic.Xtwindow(Drawing_Area));
      return;
      Loc_Set_Colour(W ,Black);
      X.Xlib.Xfillrectangle(Display => Xt.Intrinsic.Xtdisplay(Drawing_Area),
			    D => X.Drawable(Xt.Intrinsic.Xtwindow(Drawing_Area)),
			    Gc => The_Gc,
			    Xx => 0,
			    Y =>  0,
			    Width =>  Interfaces.C.Unsigned(Screen_Size_X),
			    Height => Interfaces.C.Unsigned(Screen_Size_Y));

      X.Xlib.Xsync( Display => Xt.Intrinsic.Xtdisplay(Drawing_Area),
                    Discard => 0 );
   end Blackout;

   procedure Loc_The_Drawing (W : Xt.Intrinsic.Widget) is
      Tmp : X.Signed_Int;
   begin
      Mandel_Control_Package.Generate_Plot( W => W );
   end;

   procedure ExposureCB
     (W		  : Xt.Intrinsic.Widget;
      Client_Data : Xt.Intrinsic.XtPointer;
      cbs	  : Xt.Intrinsic.XtPointer) is
   begin
--      Text_Io.Put_Line("Exposure Callback!");Text_Io.Flush;
      Blackout(W);
      Loc_The_Drawing(W);
   end;
   procedure ResizeCB
     (W		  : Xt.Intrinsic.Widget;
      Client_Data : Xt.Intrinsic.XtPointer;
      cbs	  : Xt.Intrinsic.XtPointer) is
   begin
--      Text_Io.Put_Line("Resize Callback!");Text_Io.Flush;
      null;
   end;

   First : Boolean := True;
   Zoom_Area : New_Area_Type_Package.New_Area_Type;
   Dist : Long_Integer;

   function Longer_Of (A : Long_Integer;
		       B : Long_integer) return Long_Integer is
   begin
      if A > B then 
	 return A;
      else
	 return B;
	 
      end if;
   end;

   procedure InputCB
     (W		  : Xt.Intrinsic.Widget;
      Client_Data : Xt.Intrinsic.XtPointer;
      cbs	  : Xt.Intrinsic.XtPointer) is

      type Cbs_Ptr is access Motif.Xm.XmDrawingAreaCallbackStruct;
      function To_Cbs is new 
	Unchecked_Conversion(Xt.Intrinsic.Xtpointer,
			     Cbs_ptr);
      Cb_Rec_ptr : Cbs_Ptr := To_Cbs(cbs);
      Event : X.Xlib.XButtonEvent;
   begin
--      Text_Io.Put_Line(X.Signed_Int'Image(Cb_Rec_ptr.all.Reason));
      if Cb_Rec_Ptr.all.Event.all.Event_Type = X.buttonpress then
	 Event := Cb_Rec_Ptr.all.Event.all.Xbutton;
	 Text_Io.Put_Line("(" & X.Signed_Int'Image(Event.xx) & 
			  "," & X.Signed_Int'Image(Event.Y) & ")");
	 if First then
	    Zoom_Area.Bot_Left_X := Long_Integer(Event.xx);
	    Zoom_Area.Bot_Left_y := Long_Integer(Event.y);
	    First := False;
	 else
	    Dist := Longer_Of (Long_Integer(Event.Xx) - Zoom_Area.Bot_Left_X,
			       Long_Integer(Event.Y)  - Zoom_Area.Bot_Left_Y);
	    Zoom_Area.Top_Right_X := Zoom_Area.Bot_Left_X + Dist;
	    Zoom_Area.Top_Right_Y := Zoom_Area.Bot_Left_Y + Dist;
	    First := True;
	    Constants.Set_Area(Area => New_Area_Type_Package.
			       Area_Type_Of(Zoom_Area));
	    Blackout(W);
	    Loc_The_Drawing(W);
	 end if;
      end if;
   end;
   -- Program specific callbacks
   procedure Redraw 
     (W		  : Xt.Intrinsic.Widget;
      Client_Data : Xt.Intrinsic.XtPointer;
      cbs	  : Xt.Intrinsic.XtPointer) is
   begin
      Blackout(W);
      Loc_The_Drawing(W);
   end;
   procedure Reset 
     (W		  : Xt.Intrinsic.Widget;
      Client_Data : Xt.Intrinsic.XtPointer;
      cbs	  : Xt.Intrinsic.XtPointer) is
   begin
      Constants.Set_Area;
      Blackout(W);
      Loc_The_Drawing(W);
   end;
   procedure Go_away 
     (W		  : Xt.Intrinsic.Widget;
      Client_Data : Xt.Intrinsic.XtPointer;
      cbs	  : Xt.Intrinsic.XtPointer) is

      procedure exit_prog(n: interfaces.c.int);
      pragma Import(C, exit_prog, "exit");

   begin
      Xt.Intrinsic.XtDestroywidget(My_Window);
      exit_prog(0); 
   end;
   
end;


