with Text_Io;
with Ada.Exceptions;
use  Ada.Exceptions;

with Mandel_Control_Package;

with Constants;
use  Constants;

with Complex_Type_Package;
use  Complex_Type_Package;

with Area_Queue_Package;
use  Area_Queue_Package;

with Output_Queue_Package;
use  Output_Queue_Package;

with Output_Queues;
use  Output_Queues;

with Area_Type_Package;
use  Area_Type_Package;

with New_Area_Type_Package;
use  New_Area_Type_Package;

with Queue_Array_Package;
use  Queue_Array_Package;

package body Mandel_Task_Package is

   task body Mandel_Task is

      Area    : Area_Type;

      New_Area     : New_Area_Type;
      Last_Area    : Area_Type;

      Task_Id      : Integer;

      function  Process_Point( Point : Complex ) return Colour is
	 Temp       : Complex := ( 0.0, 0.0 );
	 Iterations : Long_Integer := 0;
      begin
         while Modulus( Temp ) < 2.0 and then
               Iterations /= Max_Iterations
         loop
            Temp := Temp * Temp + Point;
            Iterations := Iterations + 1;
         end loop;
	 return Heat( Iterations );
      end Process_Point;          
      
      procedure Process_Area( Area : in Area_Type ) is

      Local_Area   : New_Area_Type;

      X_Steps      : Long_Integer := New_Area.Top_Right_X - New_Area.Bot_Left_X;
      Y_Steps      : Long_Integer := New_Area.Top_Right_Y - New_Area.Bot_Left_Y;

      New_Half_X   : Long_Integer := New_Area.Bot_Left_X + ( X_Steps ) / 2;
      New_Half_Y   : Long_Integer := New_Area.Bot_Left_Y + ( Y_Steps ) / 2;

      New_Half_Area : New_Area_Type := ( Bot_Left_X  => New_Area.Bot_Left_X,
                                         Bot_Left_Y  => New_Area.Bot_Left_Y,
                                         Top_Right_X => New_Half_X,
                                         Top_Right_Y => New_Half_Y,
                                         Top_One_Heat => False,
                                         Bot_One_Heat => False,
                                         Lef_One_Heat => False,
                                         Rig_One_Heat => False,
                                         Top_Heat     => -1,
                                         Bot_Heat     => -1,
                                         Lef_Heat     => -1,
                                         Rig_Heat     => -1 );

      Half_Area    : Area_Type     := Area_Type_Of( New_Half_Area );

      X_Half       : Float := Half_Area.Top_Right.Re;
      Y_Half       : Float := Half_Area.Top_Right.Im;

      X_Temp       : Float := Area.Bottom_Left.Re;
      Y_Temp       : Float := Area.Bottom_Left.Im;

      Last_Plot    : Float := Area.Bottom_Left.Re;

      Out_Area : New_Area_Type := New_Area;

      New_Heat     : Colour  := -1;
      Bottom_Heat  : Colour  := New_Area.Bot_Heat;
      Left_Heat    : Colour  := New_Area.Lef_Heat;
      Right_Heat   : Colour  := New_Area.Rig_Heat;
      Top_Heat     : Colour  := New_Area.Top_Heat;

      All_One_Heat    : Boolean;

      Bottom_One_Heat : Boolean := True;
      Top_One_Heat    : Boolean := True;
      Left_One_Heat   : Boolean := True;
      Right_One_Heat  : Boolean := True;

      procedure Plot_The_Bottom is
      begin
	 -- along the bottom
	 X_Temp  := Area.Bottom_Left.Re;
	 Bottom_Heat := Process_Point( Point => ( Re => X_Temp,
						  Im => Area.Bottom_Left.Im ) );
	 for Count in 0..X_Steps loop

	    New_Heat := Process_Point( Point => ( Re => X_Temp,
						  Im => Area.Bottom_Left.Im ) );
	    Bottom_One_Heat := Bottom_One_Heat and then ( New_Heat = Bottom_Heat );

	    if New_Heat /= Bottom_Heat
	    then
	       Put( Item => ( Area => 
				( Top_Right   => ( Re => X_Temp - X_Delta, Im => Area.Bottom_Left.Im ),
				  Bottom_Left => ( Re => Last_Plot,        Im => Area.Bottom_Left.Im ) ),
			      Heat => Bottom_Heat ),
                    Queue => Queue_Array( Task_Id ) );

	       Bottom_Heat := New_Heat;
	       Last_Plot := X_Temp;
	    end if;
	    X_Temp := X_Temp + X_Delta;
	 end loop;
	 Put( Item => ( Area => 
			  ( Top_Right   => ( Re => X_Temp,    Im => Area.Bottom_Left.Im ),
			    Bottom_Left => ( Re => Last_Plot, Im => Area.Bottom_Left.Im ) ),
			Heat => New_Heat ),
                    Queue => Queue_Array( Task_Id ) );
      end Plot_The_Bottom;


      procedure Plot_The_Top is
      begin
	    -- along the top
	    X_Temp  := Area.Bottom_Left.Re;         
	    Top_Heat := Process_Point( Point => ( Re => X_Temp,
						   Im => Area.Top_Right.Im ) );
	    Last_Plot := X_Temp;
	    
	    for Count in 0..X_Steps loop   
	       New_Heat := Process_Point( Point => ( Re => X_Temp,
						     Im => Area.Top_Right.Im ) );
	       Top_One_Heat := Top_One_Heat and then ( New_Heat = Top_Heat );
	       if New_Heat /= Top_Heat
	       then
		  Put( Item => ( Area => 
				   ( Top_Right   => ( Re => X_Temp - X_Delta, Im => Area.Top_Right.Im ),
				     Bottom_Left => ( Re => Last_Plot,        Im => Area.Top_Right.Im ) ),
				 Heat => Top_Heat ),
                    Queue => Queue_Array( Task_Id ) );

		  Top_Heat := New_Heat;
		  Last_Plot := X_Temp;
	       end if;
	       
	       X_Temp := X_Temp + X_Delta;
	    end loop;
	    Put( Item => ( Area => 
			     ( Top_Right   => ( Re => X_Temp,    Im => Area.Top_Right.Im ),
			       Bottom_Left => ( Re => Last_Plot, Im => Area.Top_Right.Im ) ),
			   Heat => New_Heat ),
                    Queue => Queue_Array( Task_Id ) );
      end Plot_The_Top;


      procedure Plot_The_Left is
      begin
	 -- up the left
	 Y_Temp  := Area.Bottom_Left.Im;         
	 Left_Heat := Process_Point( Point => ( Re => Area.Bottom_Left.Re,
						Im => Y_Temp ) );
	 Last_Plot := Y_Temp;
	 
	 for Count in 1..( Y_Steps - 1 ) loop
	    New_Heat := Process_Point( Point => ( Re => Area.Bottom_Left.Re,
						  Im => Y_Temp ) );
	    Left_One_Heat := Left_One_Heat and then ( New_Heat = Left_Heat );

	    if New_Heat /= Left_Heat
	    then
	       Put( Item => ( Area => 
				( Top_Right   => ( Re => Area.Bottom_Left.Re, Im => Y_Temp - Y_Delta ),
				  Bottom_Left => ( Re => Area.Bottom_Left.Re, Im => Last_Plot ) ),
			      Heat => Left_Heat ),
                    Queue => Queue_Array( Task_Id ) );

	       Left_Heat := New_Heat;
	       Last_Plot := Y_Temp;
	    end if;
	    
	    Y_Temp := Y_Temp + Y_Delta;
	 end loop;
	 Put( Item => ( Area => 
			  ( Top_Right   => ( Re => Area.Bottom_Left.Re, Im => Y_Temp ),
			    Bottom_Left => ( Re => Area.Bottom_Left.Re, Im => Last_Plot ) ),
			Heat => New_Heat ),
                    Queue => Queue_Array( Task_Id ) );
      end Plot_The_Left;

      procedure Plot_The_Right is
      begin
	 -- up the right
	 Y_Temp  := Area.Bottom_Left.Im;         
	 Right_Heat := Process_Point( Point => ( Re => Area.Top_Right.Re,
		 				 Im => Y_Temp ) );
	 Last_Plot := Y_Temp;
	 
	 for Count in 1..( Y_Steps - 1 ) loop
	    New_Heat := Process_Point( Point => ( Re => Area.Top_Right.Re,
						  Im => Y_Temp ) );
	    Right_One_Heat := Right_One_Heat and then ( New_Heat = Right_Heat );

	    if New_Heat /= Right_Heat
	    then
	       Put( Item => ( Area => 
				( Top_Right   => ( Re => Area.Top_Right.Re, Im => Y_Temp - Y_Delta ),
				  Bottom_Left => ( Re => Area.Top_Right.Re, Im => Last_Plot ) ),
			      Heat => Right_Heat ),
                    Queue => Queue_Array( Task_Id ) );

	       Right_Heat := New_Heat;
	       Last_Plot := Y_Temp;
	    end if;
	    
	    Y_Temp := Y_Temp + Y_Delta;
	 end loop;
	 Put( Item => ( Area => 
			  ( Top_Right   => ( Re => Area.Top_Right.Re, Im => Y_Temp ),
			    Bottom_Left => ( Re => Area.Top_Right.Re, Im => Last_Plot ) ),
			Heat => New_Heat ),
                    Queue => Queue_Array( Task_Id ) );
      end Plot_The_Right;


   begin


      if not New_Area.Bot_One_Heat then
         Plot_The_Bottom;
      end if;

      if not ( Y_Steps = 0 ) then
         if not New_Area.Top_One_Heat then
            Plot_The_Top;
         end if;
      else
         Top_Heat := Bottom_Heat;
      end if;

      if Y_Steps > 1 then
         if not New_Area.Lef_One_Heat then
            Plot_The_Left;
         end if;

         if not ( X_Steps = 0 ) then
            if not New_Area.Rig_One_Heat then
               Plot_The_Right;
            end if;
         else
            Right_Heat := Left_Heat;
         end if;
      else
         Left_Heat  := Bottom_Heat;
         Right_Heat := Bottom_Heat;
      end if;

      All_One_Heat := Top_One_Heat    and then
                      Bottom_One_Heat and then
                      Left_One_Heat   and then
                      Right_One_Heat  and then
                      ( Top_Heat = Bottom_Heat ) and then
                      ( Top_Heat = Left_Heat ) and then
                      ( Top_Heat = Right_Heat );

      -- what to do now?
      if All_One_Heat 
      then
	 -- paint the area we just worked out
	 Output_Queue_Package.Output_Queue.Put( Area => ( Area => Area,
    				                Heat => Bottom_Heat ) );
         -- and empty the local queue ready for next time
         Flush( Queue_Array( Task_Id ) );
      else
         -- paint the things in the output queue
         Output_Queue_Package.Output_Queue.Append( Queue_Array( Task_Id ) );

         if not Bottom_One_Heat then
            Out_Area.Bot_Left_Y := Out_Area.Bot_Left_Y + 1;
         end if;

         if not Top_One_Heat then
            Out_Area.Top_Right_Y := Out_Area.Top_Right_Y - 1;
         end if;

         if not Left_One_Heat then
            Out_Area.Bot_Left_X := Out_Area.Bot_Left_X + 1;
         end if;

         if not Right_One_Heat then
            Out_Area.Top_Right_X := Out_Area.Top_Right_X - 1;
         end if;

	 -- bottom left quadrant
	 Local_Area := ( Top_Right_X => New_Half_X,
			 Top_Right_Y => New_Half_Y,
			 Bot_Left_X  => Out_Area.Bot_Left_X,
			 Bot_Left_Y  => Out_Area.Bot_Left_Y,
                         Top_One_Heat => False,
                         Bot_One_Heat => Bottom_One_Heat,
                         Lef_One_Heat => Left_One_Heat,
                         Rig_One_Heat => False,
                         Top_Heat     => Top_Heat,
                         Bot_Heat     => Bottom_Heat,
                         Lef_Heat     => Left_Heat,
                         Rig_Heat     => Right_Heat );
	 
	 if         Local_Area.Bot_Left_X <= Local_Area.Top_Right_X
	   and then Local_Area.Bot_Left_Y <= Local_Area.Top_Right_Y
         then
	   Area_Queue.Put( Area => Local_Area );
	 end if;
	 
	 -- top left quadrant
	 Local_Area := ( Top_Right_X => New_Half_X,
			 Top_Right_Y => Out_Area.Top_Right_Y,
			 Bot_Left_X  => Out_Area.Bot_Left_X,
			 Bot_Left_Y  => New_Half_Y + 1,
                         Top_One_Heat => Top_One_Heat,
                         Bot_One_Heat => False,
                         Lef_One_Heat => Left_One_Heat,
                         Rig_One_Heat => False,
                         Top_Heat     => Top_Heat,
                         Bot_Heat     => Bottom_Heat,
                         Lef_Heat     => Left_Heat,
                         Rig_Heat     => Right_Heat );
	 
	 if         Local_Area.Bot_Left_X <= Local_Area.Top_Right_X
	   and then Local_Area.Bot_Left_Y <= Local_Area.Top_Right_Y
         then
            Area_Queue.Put( Area => Local_Area );
         end if;

         -- bottom right quadrant
         Local_Area := ( Top_Right_X => Out_Area.Top_Right_X,
                         Top_Right_Y => New_Half_Y,
                         Bot_Left_X  => New_Half_X + 1,
                         Bot_Left_Y  => Out_Area.Bot_Left_Y,
                         Top_One_Heat => False,
                         Bot_One_Heat => Bottom_One_Heat,
                         Lef_One_Heat => False,
                         Rig_One_Heat => Right_One_Heat,
                         Top_Heat     => Top_Heat,
                         Bot_Heat     => Bottom_Heat,
                         Lef_Heat     => Left_Heat,
                         Rig_Heat     => Right_Heat );

         if         Local_Area.Bot_Left_X <= Local_Area.Top_Right_X
           and then Local_Area.Bot_Left_Y <= Local_Area.Top_Right_Y
         then
            Area_Queue.Put( Area => Local_Area);
         end if;

         -- top right quadrant
         Local_Area := ( Top_Right_X => Out_Area.Top_Right_X ,
                         Top_Right_Y => Out_Area.Top_Right_Y,
                         Bot_Left_X  => New_Half_X + 1,
                         Bot_Left_Y  => New_Half_Y + 1,
                         Top_One_Heat => Top_One_Heat,
                         Bot_One_Heat => False,
                         Lef_One_Heat => False,
                         Rig_One_Heat => Right_One_Heat,
                         Top_Heat     => Top_Heat,
                         Bot_Heat     => Bottom_Heat,
                         Lef_Heat     => Left_Heat,
                         Rig_Heat     => Right_Heat );

         if         Local_Area.Bot_Left_X <= Local_Area.Top_Right_X
           and then Local_Area.Bot_Left_Y <= Local_Area.Top_Right_Y
         then
            Area_Queue.Put( Area => Local_Area);
         end if;

      end if;

   end Process_Area;

   begin

      -- wait for the first startup message before doing anything
      accept Startup( Id    : in Integer )
      do
         Task_Id := Id;
         Text_Io.Put_Line( "Task "
                         & Integer'Image( Task_Id )
                         & " started okay" );
         Text_Io.Flush;
      end Startup;

      delay 0.2;

      loop
         Area_Queue.Get( New_Area );
         Process_Area( Area_Type_Of( New_Area ) );
      end loop;           

   exception
      when E : others =>
         Text_Io.Put_Line( "mandel_task "
                         & Integer'Image( Task_Id ) );
         Text_Io.Put_Line( Exception_Name( E )); Text_Io.Flush;
         Text_Io.Put_Line( Exception_Message( E )); Text_Io.Flush;
         Text_Io.Put_Line( Exception_Information( E )); Text_Io.Flush;
         Reraise_Occurrence( E );
         
   end Mandel_Task;

end Mandel_Task_Package;


